==> superd-0.2.tar.gz: superd-0.2/internal/config/parse.go <==
// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package config

import (
	"bytes"
	"fmt"
	"io/ioutil"
	"os"
	"path/filepath"
	"regexp"
	"strings"
	"time"

	ini "github.com/subpop/go-ini"
	"github.com/trstringer/go-systemd-time/pkg/systemdtime"
)

type unit struct {
	Description string
	After       string
	Before      string
}

type service struct {
	Type          string
	ExecStart     string
	Restart       string
	RestartSec    string
	ExecStartPre  string
	ExecStartPost string
	ExecOnFailure string
	// TODO: environment field
}

type systemdConfig struct {
	Unit    unit
	Service service
}

type ErrUnsupportedValue struct {
	Option string
	Value  string
}

func (e ErrUnsupportedValue) Error() string {
	return fmt.Sprintf("unsupported value for %q: %s", e.Option, e.Value)
}

const (
	// systemd defaults to 100ms, but that's too fast...
	defaultRestartDur = 1 * time.Second
)

func replaceInstanceName(contents []byte, name string) (out []byte) {
	out = bytes.ReplaceAll(contents, []byte("%i"), []byte(name))
	return
}

// Parse the "unit" name from the path.
// Expects the path to be to a 'full' instanced template file, e.g.
// /path/to/foo@bar.service and *not* /path/to/foo@.service
// Returns the instance name, or an empty string if unable to find it
func instanceName(path string) (name string) {
	file := filepath.Base(strings.TrimSuffix(path, ".service"))
	if strings.HasPrefix(file, "@") {
		return
	}

	parts := strings.Split(file, "@")
	if len(parts) != 2 {
		return
	}

	name = parts[1]
	return
}

func Parse(path string) (s *ServiceConfig, err error) {
	// TODO: this ignore escaping of "unit names"... so %i is basically %I
	// TODO: also implement other specifiers?
	// see: https://www.freedesktop.org/software/systemd/man/systemd.unit.html

	template, err := regexp.MatchString(".+@.+.service", filepath.Base(path))
	if err != nil {
		err = fmt.Errorf("config.Parse: %w", err)
		return
	}

	contents, err := ioutil.ReadFile(path)
	if err != nil {
		err = fmt.Errorf("config.Parse: %w", err)
		return
	}

	if template {
		contents = replaceInstanceName(contents, instanceName(path))
	}

	sysd := systemdConfig{}

	o := ini.Options{AllowNumberSignComments: true}

	if err = ini.UnmarshalWithOptions(contents, &sysd, o); err != nil {
		err = fmt.Errorf("config.Parse: %w", err)
		return
	}

	name := strings.Split(filepath.Base(path), filepath.Ext(path))[0]

	s, err = parseConfig(name, &sysd)
	if err != nil {
		return
	}

	s.Path = path

	return
}

func parseDeps(s string) (deps []string, err error) {

	for _, dep := range strings.Fields(s) {
		if dep == "" {
			continue
		}
		// Sometimes dependencies are listed with a ".service" extension, those
		// are allowed here, but any other is an error since we don't want to
		// allow things that aren't services (e.g. a ".socket", if that's even
		// valid..)
		if ext := filepath.Ext(dep); ext != "" {
			if strings.ToLower(ext) == ".service" {
				dep = strings.TrimSuffix(dep, ext)
			} else {
				err = ErrUnsupportedValue{Option: "Before=|After=", Value: s}
				fmt.Println(err)
				return
			}
		}
		deps = append(deps, dep)
	}
	return
}

func parseCmd(s string, allowEmpty bool) (cmd string, args []string, err error) {
	// split ExecStart into cmd string + args slice
	execFields := strings.Fields(s)
	if len(execFields) < 1 {
		if !allowEmpty {
			err = ErrUnsupportedValue{Option: "ExecStart=", Value: s}
		}
		return
	}

	cmd = execFields[0]
	if len(execFields) > 1 {
		args = execFields[1:]
	}
	// expand any env vars in the arguments. this loosely mimicks systemd's
	// "basic" variable expansion support, and probably needs to be revisited
	// at some point. TODO!
	for i := range args {
		args[i] = os.ExpandEnv(args[i])
	}

	return
}

func parseRestart(s string) (restart RestartCond, err error) {

	switch strings.ToUpper(s) {
	case "ALWAYS":
		restart = RestartAlways
	case "ON-FAILURE":
		restart = RestartFailure
	case "ON-SUCCESS":
		restart = RestartSuccess
	case "NO":
		restart = RestartNever
	default:
		err = ErrUnsupportedValue{Option: "Restart=", Value: s}
	}

	return
}

func parseRestartDur(s string) (restartDur time.Duration, err error) {
	// a missing value for this option is not fatal
	if s != "" {
		if restartDur, err = systemdtime.ToDuration(s); err != nil {
			// assume it's in seconds and try again
			if restartDur, err = systemdtime.ToDuration(s + "s"); err != nil {
				err = ErrUnsupportedValue{Option: "RestartSec=", Value: s}
			}
		}
	} else {
		restartDur = defaultRestartDur
	}

	return
}

func parseType(s string) (t Type, err error) {
	switch strings.ToUpper(s) {
	case "":
	case "SIMPLE":
		t = TypeSimple
	case "ONESHOT":
		t = TypeOneshot
	default:
		err = ErrUnsupportedValue{Option: "Type=", Value: s}
	}

	return
}

func parseConfig(name string, sysd *systemdConfig) (s *ServiceConfig, err error) {
	s = &ServiceConfig{
		Name:       name,
		RestartDur: defaultRestartDur,
	}

	// "Before" dependencies
	if s.Before, err = parseDeps(sysd.Unit.Before); err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// "After" dependencies
	if s.After, err = parseDeps(sysd.Unit.After); err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// ExecStart=
	s.Cmd, s.CmdArgs, err = parseCmd(sysd.Service.ExecStart, false)
	if err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// ExecStartPre=
	s.ExecStartPreCmd, s.ExecStartPreCmdArgs, err = parseCmd(sysd.Service.ExecStartPre, true)
	if err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// ExecStartPost=
	s.ExecStartPostCmd, s.ExecStartPostCmdArgs, err = parseCmd(sysd.Service.ExecStartPost, true)
	if err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// ExecOnFailure=
	s.ExecOnFailureCmd, s.ExecOnFailureCmdArgs, err = parseCmd(sysd.Service.ExecOnFailure, true)
	if err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// Restart=
	if s.Restart, err = parseRestart(sysd.Service.Restart); err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// RestartSec=
	if s.RestartDur, err = parseRestartDur(sysd.Service.RestartSec); err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// Type=
	if s.Type, err = parseType(sysd.Service.Type); err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	return
}

==> superd-0.3.2.tar.gz: superd-0.3.2/internal/config/parse.go <==
// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package config

import (
	"bytes"
	"fmt"
	"io/ioutil"
	"os"
	"path/filepath"
	"regexp"
	"strings"
	"time"

	ini "git.sr.ht/~spc/go-ini"
	"github.com/trstringer/go-systemd-time/pkg/systemdtime"
)

type unit struct {
	Description           string
	After                 string
	Before                string
	StartLimitIntervalSec string
	StartLimitBurst       int
}

type service struct {
	Type          string
	ExecStart     string
	Restart       string
	RestartSec    string
	ExecStartPre  string
	ExecStartPost string
	ExecOnFailure string
	// TODO: environment field
}

type systemdConfig struct {
	Unit    unit
	Service service
}

type ErrUnsupportedValue struct {
	Option string
	Value  string
}

func (e ErrUnsupportedValue) Error() string {
	return fmt.Sprintf("unsupported value for %q: %s", e.Option, e.Value)
}

const (
	// copied from systemd defaults
	defaultRestartDur      = 100 * time.Millisecond
	defaultStartLimitDur   = 10 * time.Second
	defaultStartLimitBurst = 5
)

func replaceInstanceName(contents []byte, name string) (out []byte) {
	out = bytes.ReplaceAll(contents, []byte("%i"), []byte(name))
	return
}

func multiline2single(s string) string {
	out := strings.TrimSpace(s)
	out = strings.ReplaceAll(out, "\n", "")
	out = strings.ReplaceAll(out, "\r", "")
	return strings.ReplaceAll(out, "\\", "")
}

// Parse the "unit" name from the path.
// Expects the path to be to a 'full' instanced template file, e.g.
// /path/to/foo@bar.service and *not* /path/to/foo@.service
// Returns the instance name, or an empty string if unable to find it
func instanceName(path string) (name string) {
	file := filepath.Base(strings.TrimSuffix(path, ".service"))
	if strings.HasPrefix(file, "@") {
		return
	}

	parts := strings.Split(file, "@")
	if len(parts) != 2 {
		return
	}

	name = parts[1]
	return
}

func Parse(path string) (s *ServiceConfig, err error) {
	// TODO: this ignore escaping of "unit names"... so %i is basically %I
	// TODO: also implement other specifiers?
	// see: https://www.freedesktop.org/software/systemd/man/systemd.unit.html

	template, err := regexp.MatchString(".+@.+.service", filepath.Base(path))
	if err != nil {
		err = fmt.Errorf("config.Parse: %w", err)
		return
	}

	contents, err := ioutil.ReadFile(path)
	if err != nil {
		err = fmt.Errorf("config.Parse: %w", err)
		return
	}

	if template {
		contents = replaceInstanceName(contents, instanceName(path))
	}

	sysd := systemdConfig{}

	o := ini.Options{
		AllowNumberSignComments: true,
		AllowMultilineValues:    true,
		AllowEmptyValues:        true,
	}

	if err = ini.UnmarshalWithOptions(contents, &sysd, o); err != nil {
		err = fmt.Errorf("config.Parse: %w", err)
		return
	}

	name := strings.Split(filepath.Base(path), filepath.Ext(path))[0]

	s, err = parseConfig(name, &sysd)
	if err != nil {
		return
	}

	s.Path = path

	return
}

func parseDeps(s string) (deps []string, err error) {

	for _, dep := range strings.Fields(strings.TrimSpace(s)) {
		if dep == "" {
			continue
		}
		// Sometimes dependencies are listed with a ".service" extension, those
		// are allowed here, but any other is an error since we don't want to
		// allow things that aren't services (e.g. a ".socket", if that's even
		// valid..)
		if ext := filepath.Ext(dep); ext != "" {
			if strings.ToLower(ext) == ".service" {
				dep = strings.TrimSuffix(dep, ext)
			} else {
				err = ErrUnsupportedValue{Option: "Before=|After=", Value: s}
				fmt.Println(err)
				return
			}
		}
		deps = append(deps, dep)
	}
	return
}

func parseCmd(s string, allowEmpty bool) (cmd string, args []string, err error) {
	// split ExecStart into cmd string + args slice
	execFields := strings.Fields(multiline2single(s))
	if len(execFields) < 1 {
		if !allowEmpty {
			err = ErrUnsupportedValue{Option: "ExecStart=", Value: s}
		}
		return
	}

	cmd = execFields[0]
	if len(execFields) > 1 {
		args = execFields[1:]
	}
	// expand any env vars in the arguments. this loosely mimicks systemd's
	// "basic" variable expansion support, and probably needs to be revisited
	// at some point. TODO!
	for i := range args {
		args[i] = os.ExpandEnv(args[i])
	}

	return
}

func parseRestart(s string) (restart RestartCond, err error) {

	switch strings.ToUpper(strings.TrimSpace(s)) {
	case "ALWAYS":
		restart = RestartAlways
	case "ON-FAILURE":
		restart = RestartFailure
	case "ON-SUCCESS":
		restart = RestartSuccess
	case "NO":
		restart = RestartNever
	default:
		err = ErrUnsupportedValue{Option: "Restart=", Value: s}
	}

	return
}

func parseDur(s string, optionName string, required bool, defaultDur time.Duration) (dur time.Duration, err error) {

	if required && s == "" {
		err = ErrUnsupportedValue{Option: optionName, Value: "value required but none was given"}
		return
	}

	if s != "" {
		if dur, err = systemdtime.ToDuration(strings.TrimSpace(s)); err != nil {
			// assume it's in seconds and try again
			if dur, err = systemdtime.ToDuration(s + "s"); err != nil {
				err = ErrUnsupportedValue{Option: optionName, Value: s}
			}
		}
	} else {
		dur = defaultDur
	}

	return
}

func parseType(s string) (t Type, err error) {
	switch strings.ToUpper(strings.TrimSpace(s)) {
	case "":
	case "SIMPLE":
		t = TypeSimple
	case "ONESHOT":
		t = TypeOneshot
	default:
		err = ErrUnsupportedValue{Option: "Type=", Value: s}
	}

	return
}

func parseConfig(name string, sysd *systemdConfig) (s *ServiceConfig, err error) {
	s = &ServiceConfig{
		Name:       name,
		RestartDur: defaultRestartDur,
	}

	// Description=
	s.Description = multiline2single(sysd.Unit.Description)

	// "Before" dependencies
	if s.Before, err = parseDeps(sysd.Unit.Before); err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// "After" dependencies
	if s.After, err = parseDeps(sysd.Unit.After); err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// ExecStart=
	s.Cmd, s.CmdArgs, err = parseCmd(multiline2single(sysd.Service.ExecStart), false)
	if err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// StartLimitIntervalSec=
	if s.StartLimitDur, err = parseDur(sysd.Unit.StartLimitIntervalSec, "StartLimitIntervalSec", false, defaultStartLimitDur); err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// StartLimitBurst=
	s.StartLimitBurst = sysd.Unit.StartLimitBurst
	if s.StartLimitBurst == 0 {
		s.StartLimitBurst = defaultStartLimitBurst
	}

	// ExecStartPre=
	s.ExecStartPreCmd, s.ExecStartPreCmdArgs, err = parseCmd(sysd.Service.ExecStartPre, true)
	if err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// ExecStartPost=
	s.ExecStartPostCmd, s.ExecStartPostCmdArgs, err = parseCmd(sysd.Service.ExecStartPost, true)
	if err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// ExecOnFailure=
	s.ExecOnFailureCmd, s.ExecOnFailureCmdArgs, err = parseCmd(sysd.Service.ExecOnFailure, true)
	if err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// Restart=
	if s.Restart, err = parseRestart(sysd.Service.Restart); err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// RestartSec=
	if s.RestartDur, err = parseDur(sysd.Service.RestartSec, "Restart=", false, defaultRestartDur); err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// Type=
	if s.Type, err = parseType(sysd.Service.Type); err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	return
}
